/* ipuz-style.c
 *
 * Copyright 2022 Jonathan Blandford <jrb@gnome.org>
 *
 * This file is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 3 of the
 * License, or (at your option) any later version.
 *
 * This file is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * SPDX-License-Identifier: LGPL-3.0-or-later
 */
#include "ipuz-style.h"


struct _IPuzStyle
{
  grefcount ref_count;
  gchar *style_name; /* Internal name of the style, for shared styles */

  IPuzStyleShapeEnum shapebg;
  gboolean highlight;
  gchar *named;
  gint border;
  IPuzStyleDivided divided;
  gchar *label;
  GHashTable *mark; /* Hashtable of strings indexed by IPUZ_STYLE_MARK_ */
  gchar *imagebg_url;
  gchar *image_url;
  gchar *bg_color;
  gchar *text_color;
  gchar *border_color;
  IPuzStyleSides barred;
  IPuzStyleSides dotted;
  IPuzStyleSides dashed;
  IPuzStyleSides lessthan;
  IPuzStyleSides greaterthan;
  IPuzStyleSides equal;
};


G_DEFINE_BOXED_TYPE (IPuzStyle, ipuz_style, ipuz_style_ref, ipuz_style_unref);


static IPuzStyleShapeEnum  parse_shapebg         (const char         *shapebg);
static IPuzStyleSides      parse_sides           (const char         *val);
static char               *sides_to_str          (IPuzStyleSides      sides);
static const char         *shapebg_to_str        (IPuzStyleShapeEnum  shapebg);
static char               *parse_color           (JsonNode           *node);
static void                ipuz_style_parse_mark (IPuzStyle          *style,
                                                  JsonNode           *node);

IPuzStyle *
ipuz_style_new (void)
{
  IPuzStyle *style;

  style = (IPuzStyle *) g_new0 (IPuzStyle, 1);
  g_ref_count_init (&style->ref_count);

  return style;
}

static IPuzStyleSides
sides_from_json (JsonObject *obj, const char *prop_name)
{
  JsonNode *element = json_object_get_member (obj, prop_name);
  if (element)
    {
      const gchar *val = json_node_get_string (element);
      return parse_sides (val);
    }
  else
    {
      return 0;
    }
}

IPuzStyle *
ipuz_style_new_from_json (JsonNode *node)
{
  IPuzStyle *style;

  g_return_val_if_fail (node != NULL, NULL);

  style = ipuz_style_new ();

  if (JSON_NODE_HOLDS_OBJECT (node))
    {
      JsonObject *obj;
      JsonNode *element;

      obj = json_node_get_object (node);

      element = json_object_get_member (obj, "shapebg");
      if (element)
        {
          const gchar *shapebg = json_node_get_string (element);
          style->shapebg = parse_shapebg (shapebg);
        }

      element = json_object_get_member (obj, "highlight");
      if (element)
        style->highlight = json_node_get_boolean (element);

      element = json_object_get_member (obj, "named");
      if (element)
        style->named = g_strdup (json_node_get_string (element));

      element = json_object_get_member (obj, "border");
      if (element)
        style->border = json_node_get_int (element);

      element = json_object_get_member (obj, "divided");
      if (element)
        {
          const gchar *divided = json_node_get_string (element);

          if (g_strcmp0 (divided, "-") == 0)
            style->divided = IPUZ_STYLE_DIVIDED_HORIZ;
          else if (g_strcmp0 (divided, "|") == 0)
            style->divided = IPUZ_STYLE_DIVIDED_VERT;
          else if (g_strcmp0 (divided, "/") == 0)
            style->divided = IPUZ_STYLE_DIVIDED_UP_RIGHT;
          else if (g_strcmp0 (divided, "\\") == 0)
            style->divided = IPUZ_STYLE_DIVIDED_UP_LEFT;
          else if (g_strcmp0 (divided, "+") == 0)
            style->divided = IPUZ_STYLE_DIVIDED_PLUS;
          else if (g_strcmp0 (divided, "x") == 0)
            style->divided = IPUZ_STYLE_DIVIDED_CROSS;
          else
            style->divided = IPUZ_STYLE_DIVIDED_NONE;
        }

      element = json_object_get_member (obj, "label");
      if (element)
        style->label = g_strdup (json_node_get_string (element));

      element = json_object_get_member (obj, "mark");
      if (element)
        ipuz_style_parse_mark (style, element);

      element = json_object_get_member (obj, "image");
      if (element)
        style->image_url = g_strdup (json_node_get_string (element));

      element = json_object_get_member (obj, "imagebg");
      if (element)
        style->imagebg_url = g_strdup (json_node_get_string (element));

      element = json_object_get_member (obj, "color");
      if (element)
          style->bg_color = parse_color (element);

      element = json_object_get_member (obj, "colortext");
      if (element)
          style->text_color = parse_color (element);

      element = json_object_get_member (obj, "colorborder");
      if (element)
          style->border_color = parse_color (element);

      style->barred      = sides_from_json (obj, "barred");
      style->dotted      = sides_from_json (obj, "dotted");
      style->dashed      = sides_from_json (obj, "dashed");
      style->lessthan    = sides_from_json (obj, "lessthan");
      style->greaterthan = sides_from_json (obj, "greaterthan");
      style->equal       = sides_from_json (obj, "equal");
    }

  return style;
}

IPuzStyle *
ipuz_style_ref (IPuzStyle *style)
{
  g_return_val_if_fail (style != NULL, NULL);

  g_ref_count_inc (&style->ref_count);

  return style;
}

void
ipuz_style_unref (IPuzStyle *style)
{
  if (style == NULL)
    return;

  if (!g_ref_count_dec (&style->ref_count))
    return;

  /* Free */
  if (style->mark)
    g_hash_table_unref (style->mark);

  g_free (style->style_name);
  g_free (style->named);
  g_free (style->label);
  g_free (style->image_url);
  g_free (style->imagebg_url);
  g_free (style->bg_color);
  g_free (style->text_color);
  g_free (style->border_color);

  g_free (style);
}


static void
mark_copy_func (gpointer    key,
                const char *mark_label,
                GHashTable *dest_marks)
{
  g_hash_table_insert (dest_marks, key, g_strdup (mark_label));
}

IPuzStyle *
ipuz_style_copy (IPuzStyle *style)
{
  IPuzStyle *new_style;

  if (style == NULL)
    return NULL;

  new_style = ipuz_style_new ();
  new_style->style_name = g_strdup (style->style_name);

  new_style->named = g_strdup (style->named);
  new_style->border = style->border;
  new_style->shapebg = style->shapebg;
  new_style->highlight = style->highlight;
  new_style->divided = style->divided;
  if (style->mark)
    {
      new_style->mark = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, (GDestroyNotify) g_free);
      g_hash_table_foreach (style->mark, (GHFunc) mark_copy_func, new_style->mark);
    }
  new_style->label = g_strdup (style->label);
  new_style->image_url = g_strdup (style->image_url);
  new_style->imagebg_url = g_strdup (style->imagebg_url);
  new_style->bg_color = g_strdup (style->bg_color);
  new_style->text_color = g_strdup (style->text_color);
  new_style->border_color = g_strdup (style->border_color);
  new_style->barred = style->barred;
  new_style->dotted = style->dotted;
  new_style->dashed = style->dashed;
  new_style->lessthan = style->lessthan;
  new_style->greaterthan = style->greaterthan;
  new_style->equal = style->equal;

  return new_style;
}


static void
build_mark_foreach_cb (IPuzStyleMark  mark,
                       const gchar   *label,
                       gpointer       user_data)

{
  JsonBuilder *builder = JSON_BUILDER (user_data);

  switch (mark)
    {
    case IPUZ_STYLE_MARK_TL:
      json_builder_set_member_name (builder, "TL");
      break;
    case IPUZ_STYLE_MARK_T:
      json_builder_set_member_name (builder, "T");
      break;
    case IPUZ_STYLE_MARK_TR:
      json_builder_set_member_name (builder, "TR");
      break;
    case IPUZ_STYLE_MARK_L:
      json_builder_set_member_name (builder, "L");
      break;
    case IPUZ_STYLE_MARK_C:
      json_builder_set_member_name (builder, "C");
      break;
    case IPUZ_STYLE_MARK_R:
      json_builder_set_member_name (builder, "R");
      break;
    case IPUZ_STYLE_MARK_BL:
      json_builder_set_member_name (builder, "BL");
      break;
    case IPUZ_STYLE_MARK_B:
      json_builder_set_member_name (builder, "B");
      break;
    case IPUZ_STYLE_MARK_BR:
      json_builder_set_member_name (builder, "BR");
      break;
    default:
      g_warning ("STYLE as a bitfield is not supported");
      json_builder_set_member_name (builder, "C");
      break;
    }
  json_builder_add_string_value (builder, label);
}

static void
sides_to_json (JsonBuilder *builder, const char *prop_name, IPuzStyleSides sides)
{
  if (sides != 0)
    {
      char *val = sides_to_str (sides);
      json_builder_set_member_name (builder, prop_name);
      json_builder_add_string_value (builder, val);
      g_free (val);
    }
}

void
ipuz_style_build (IPuzStyle   *style,
                  JsonBuilder *builder)
{
  g_return_if_fail (style != NULL);

  json_builder_begin_object (builder);


  if (style->shapebg != IPUZ_STYLE_SHAPE_NONE)
    {
      json_builder_set_member_name (builder, "shapebg");
      json_builder_add_string_value (builder, shapebg_to_str (style->shapebg));
    }

  /* FIXME(parse): add highlight_set, or an enum */
  if (style->highlight)
    {
      json_builder_set_member_name (builder, "highlight");
      json_builder_add_boolean_value (builder, style->highlight);
    }

  if (style->named)
    {
      json_builder_set_member_name (builder, "named");
      json_builder_add_string_value (builder, style->named);
    }

  /* FIXME(build): border needs a border-set variable. 0 is a valid value */
  if (style->border != 0)
    {
      json_builder_set_member_name (builder, "border");
      json_builder_add_int_value (builder, style->border);
    }

  if (style->divided != IPUZ_STYLE_DIVIDED_NONE)
    {
      json_builder_set_member_name (builder, "divided");
      switch (style->divided)
        {
        case IPUZ_STYLE_DIVIDED_HORIZ:
          json_builder_add_string_value (builder, "-");
          break;
        case IPUZ_STYLE_DIVIDED_VERT:
          json_builder_add_string_value (builder, "|");
          break;
        case IPUZ_STYLE_DIVIDED_UP_RIGHT:
          json_builder_add_string_value (builder, "/");
          break;
        case IPUZ_STYLE_DIVIDED_UP_LEFT:
          json_builder_add_string_value (builder, "\\");
          break;
        case IPUZ_STYLE_DIVIDED_PLUS:
          json_builder_add_string_value (builder, "+");
          break;
        case IPUZ_STYLE_DIVIDED_CROSS:
          json_builder_add_string_value (builder, "X");
          break;
        default:
          g_warning ("unknown divided style");
          json_builder_add_string_value (builder, "?");
          break;
        }
    }

  if (style->label)
    {
      json_builder_set_member_name (builder, "label");
      json_builder_add_string_value (builder, style->label);
    }

  if (style->mark)
    {
      json_builder_set_member_name (builder, "mark");
      json_builder_begin_object (builder);
      ipuz_style_mark_foreach (style, build_mark_foreach_cb, builder);
      json_builder_end_object (builder);
    }

  if (style->imagebg_url)
    {
      json_builder_set_member_name (builder, "imagebg");
      json_builder_add_string_value (builder, style->imagebg_url);
    }

  if (style->image_url)
    {
      json_builder_set_member_name (builder, "label");
      json_builder_add_string_value (builder, style->image_url);

    }

  if (style->bg_color)
      {
        json_builder_set_member_name (builder, "color");
        json_builder_add_string_value (builder, style->bg_color);
      }

  if (style->text_color)
        {
          json_builder_set_member_name (builder, "colortext");
          json_builder_add_string_value (builder, style->text_color);
        }

  if (style->border_color)
    {
      json_builder_set_member_name (builder, "colorborder");
      json_builder_add_string_value (builder, style->border_color);
    }

  sides_to_json (builder, "barred",      style->barred);
  sides_to_json (builder, "dotted",      style->dotted);
  sides_to_json (builder, "dashed",      style->dashed);
  sides_to_json (builder, "lessthan",    style->lessthan);
  sides_to_json (builder, "greaterthan", style->greaterthan);
  sides_to_json (builder, "equal",       style->equal);

  json_builder_end_object (builder);
}

void
ipuz_style_set_style_name (IPuzStyle   *style,
                           const gchar *style_name)
{
  g_return_if_fail (style != NULL);

  g_clear_pointer (&style->style_name, g_free);
  style->style_name = g_strdup (style_name);
}

/* Getters */
const gchar *
ipuz_style_get_style_name (IPuzStyle *style)
{
  g_return_val_if_fail (style != NULL, NULL);

  return style->style_name;
}

const gchar *
ipuz_style_get_named (IPuzStyle *style)
{
  g_return_val_if_fail (style != NULL, NULL);

  return style->named;
}

gint
ipuz_style_get_border (IPuzStyle *style)
{
  g_return_val_if_fail (style != NULL, 0);

  return style->border;
}

IPuzStyleShapeEnum
ipuz_style_get_shapebg (IPuzStyle *style)
{
  g_return_val_if_fail (style != NULL, IPUZ_STYLE_SHAPE_NONE);

  return style->shapebg;
}

gboolean
ipuz_style_get_highlight (IPuzStyle *style)
{
  g_return_val_if_fail (style != NULL, FALSE);

  return style->highlight;
}

IPuzStyleDivided
ipuz_style_get_divided (IPuzStyle *style)
{
  g_return_val_if_fail (style != NULL, IPUZ_STYLE_DIVIDED_NONE);

  return style->divided;
}

typedef struct
{
  IPuzMarkFunc *func;
  gpointer user_data;
} ForeachTuple;


void
mark_foreach (gpointer key,
              gpointer value,
              gpointer user_data)
{
  ForeachTuple *tuple = (ForeachTuple *)user_data;

  (*tuple->func) (GPOINTER_TO_INT (key), (const char *) value, tuple->user_data);
}

void
ipuz_style_mark_foreach (IPuzStyle    *style,
                         IPuzMarkFunc  func,
                         gpointer      user_data)
{
  ForeachTuple tuple;

  g_return_if_fail (style != NULL);
  g_return_if_fail (func != NULL);

  if (style->mark)
    {
      tuple.func = &func;
      tuple.user_data = user_data;

      g_hash_table_foreach (style->mark, mark_foreach, &tuple);
    }
}

const gchar *
ipuz_style_get_label (IPuzStyle *style)
{
  g_return_val_if_fail (style != NULL, NULL);

  return style->label;
}

const gchar *
ipuz_style_get_image_url (IPuzStyle *style)
{
  g_return_val_if_fail (style != NULL, NULL);

  return style->image_url;
}

const gchar *
ipuz_style_get_imagebg_url (IPuzStyle *style)
{
  g_return_val_if_fail (style != NULL, NULL);

  return style->imagebg_url;
}

const gchar *
ipuz_style_get_bg_color (IPuzStyle *style)
{
  g_return_val_if_fail (style != NULL, NULL);

  return style->bg_color;
}

const gchar *
ipuz_style_get_text_color (IPuzStyle *style)
{
  g_return_val_if_fail (style != NULL, NULL);

  return style->text_color;
}

const gchar *
ipuz_style_get_border_color (IPuzStyle *style)
{
  g_return_val_if_fail (style != NULL, NULL);

  return style->border_color;
}

IPuzStyleSides
ipuz_style_get_barred (IPuzStyle *style)
{
  g_return_val_if_fail (style != NULL, 0);

  return style->barred;
}

IPuzStyleSides
ipuz_style_get_dotted (IPuzStyle *style)
{
  g_return_val_if_fail (style != NULL, 0);

  return style->dotted;
}

IPuzStyleSides
ipuz_style_get_dashed (IPuzStyle *style)
{
  g_return_val_if_fail (style != NULL, 0);

  return style->dashed;
}

IPuzStyleSides
ipuz_style_get_lessthan (IPuzStyle *style)
{
  g_return_val_if_fail (style != NULL, 0);

  return style->lessthan;
}

IPuzStyleSides
ipuz_style_get_greaterthan (IPuzStyle *style)
{
  g_return_val_if_fail (style != NULL, 0);

  return style->greaterthan;
}

IPuzStyleSides
ipuz_style_get_equal (IPuzStyle *style)
{
  g_return_val_if_fail (style != NULL, 0);

  return style->equal;
}


static char *
parse_color (JsonNode *node)
{
  GValue value = G_VALUE_INIT;
  gchar *retval = NULL;

  json_node_get_value (node, &value);
  if (G_VALUE_HOLDS_STRING (&value))
    {
      retval = g_value_dup_string (&value);
    }
  else if (G_VALUE_HOLDS_INT (&value) || G_VALUE_HOLDS_INT64 (&value))
    {
      int number = json_node_get_int (node);
      /* FIXME: pick a pallet */
      if (number == 0)
        retval = g_strdup ("black");
      else
        retval = NULL;
    }
  g_value_unset (&value);

  return retval;
}

static void
ipuz_style_parse_mark (IPuzStyle *style,
                       JsonNode  *node)
{
  if (style->mark == NULL)
    style->mark = g_hash_table_new_full (g_direct_hash, g_direct_equal, NULL, (GDestroyNotify) g_free);

  if (JSON_NODE_HOLDS_OBJECT (node))
    {
      const gchar *mark;
      JsonObject *obj;
      JsonNode *element;

      obj = json_node_get_object (node);

      element = json_object_get_member (obj, "TL");
      if (element)
        {
          mark = json_node_get_string (element);
          g_hash_table_insert (style->mark,
                               GINT_TO_POINTER (IPUZ_STYLE_MARK_TL),
                               g_strdup (mark));
          }
      element = json_object_get_member (obj, "T");
      if (element)
        {
          mark = json_node_get_string (element);
          g_hash_table_insert (style->mark,
                               GINT_TO_POINTER (IPUZ_STYLE_MARK_T),
                               g_strdup (mark));
          }
      element = json_object_get_member (obj, "TR");
      if (element)
        {
          mark = json_node_get_string (element);
          g_hash_table_insert (style->mark,
                               GINT_TO_POINTER (IPUZ_STYLE_MARK_TR),
                               g_strdup (mark));
          }
      element = json_object_get_member (obj, "L");
      if (element)
        {
          mark = json_node_get_string (element);
          g_hash_table_insert (style->mark,
                               GINT_TO_POINTER (IPUZ_STYLE_MARK_L),
                               g_strdup (mark));
          }
      element = json_object_get_member (obj, "C");
      if (element)
        {
          mark = json_node_get_string (element);
          g_hash_table_insert (style->mark,
                               GINT_TO_POINTER (IPUZ_STYLE_MARK_C),
                               g_strdup (mark));
          }
      element = json_object_get_member (obj, "R");
      if (element)
        {
          mark = json_node_get_string (element);
          g_hash_table_insert (style->mark,
                               GINT_TO_POINTER (IPUZ_STYLE_MARK_R),
                               g_strdup (mark));
          }
      element = json_object_get_member (obj, "BL");
      if (element)
        {
          mark = json_node_get_string (element);
          g_hash_table_insert (style->mark,
                               GINT_TO_POINTER (IPUZ_STYLE_MARK_BL),
                               g_strdup (mark));
          }
      element = json_object_get_member (obj, "B");
      if (element)
        {
          mark = json_node_get_string (element);
          g_hash_table_insert (style->mark,
                               GINT_TO_POINTER (IPUZ_STYLE_MARK_B),
                               g_strdup (mark));
        }
      element = json_object_get_member (obj, "BR");
      if (element)
        {
          mark = json_node_get_string (element);
          g_hash_table_insert (style->mark,
                               GINT_TO_POINTER (IPUZ_STYLE_MARK_BR),
                               g_strdup (mark));
        }

      }
  /* Make sure we actually added something */
  if (g_hash_table_size (style->mark) == 0)
    {
      g_hash_table_unref (style->mark);
      style->mark = NULL;
    }
}

struct shapebg_name {
  const char *shapebg;
  IPuzStyleShapeEnum value;
};

static const struct shapebg_name shapebg_names[] = {
  { "circle",         IPUZ_STYLE_SHAPE_CIRCLE },
  { "arrow-left",     IPUZ_STYLE_SHAPE_ARROW_LEFT },
  { "arrow-right",    IPUZ_STYLE_SHAPE_ARROW_RIGHT },
  { "arrow-up",       IPUZ_STYLE_SHAPE_ARROW_UP },
  { "arrow-down",     IPUZ_STYLE_SHAPE_ARROW_DOWN },
  { "triangle-left",  IPUZ_STYLE_SHAPE_TRIANGLE_LEFT },
  { "triangle-right", IPUZ_STYLE_SHAPE_TRIANGLE_RIGHT },
  { "triangle-up",    IPUZ_STYLE_SHAPE_TRIANGLE_UP },
  { "triangle-down",  IPUZ_STYLE_SHAPE_TRIANGLE_DOWN },
  { "diamond",        IPUZ_STYLE_SHAPE_DIAMOND },
  { "club",           IPUZ_STYLE_SHAPE_CLUB },
  { "heart",          IPUZ_STYLE_SHAPE_HEART },
  { "spade",          IPUZ_STYLE_SHAPE_SPADE },
  { "star",           IPUZ_STYLE_SHAPE_STAR },
  { "square",         IPUZ_STYLE_SHAPE_SQUARE },
  { "rhombus",        IPUZ_STYLE_SHAPE_RHOMBUS },
  { "/",              IPUZ_STYLE_SHAPE_SLASH },
  { "\\",             IPUZ_STYLE_SHAPE_BACKSLASH },
  { "X",              IPUZ_STYLE_SHAPE_X },
};

static IPuzStyleShapeEnum
parse_shapebg (const char *shapebg)
{
  guint i;

  for (i = 0; i < G_N_ELEMENTS (shapebg_names); i++)
    {
      if (g_strcmp0 (shapebg, shapebg_names[i].shapebg) == 0)
        return shapebg_names[i].value;
    }

  return IPUZ_STYLE_SHAPE_NONE;
}

static const char *
shapebg_to_str (IPuzStyleShapeEnum shapebg)
{
  guint i;

  for (i = 0; i < G_N_ELEMENTS (shapebg_names); i++)
    {
      if (shapebg == shapebg_names[i].value)
        return shapebg_names[i].shapebg;
    }

  g_assert_not_reached ();
}

static IPuzStyleSides
parse_sides (const char *val)
{
  IPuzStyleSides sides = 0;
  const char *p;

  for (p = val; *p; p++)
    {
      switch (*p) {
      case 'T':
        sides |= IPUZ_STYLE_SIDES_TOP;
        break;

      case 'R':
        sides |= IPUZ_STYLE_SIDES_RIGHT;
        break;

      case 'B':
        sides |= IPUZ_STYLE_SIDES_BOTTOM;
        break;

      case 'L':
        sides |= IPUZ_STYLE_SIDES_LEFT;
        break;

      default:
        break;
      }
    }

  return sides;
}

static char *
sides_to_str (IPuzStyleSides sides)
{
  char val[5] = { 0 };
  gsize i = 0;

  if (sides & IPUZ_STYLE_SIDES_TOP)
    {
      val[i] = 'T';
      i++;
    }

  if (sides & IPUZ_STYLE_SIDES_RIGHT)
    {
      val[i] = 'R';
      i++;
    }

  if (sides & IPUZ_STYLE_SIDES_BOTTOM)
    {
      val[i] = 'B';
      i++;
    }

  if (sides & IPUZ_STYLE_SIDES_LEFT)
    {
      val[i] = 'L';
      i++;
    }

  g_assert (i < G_N_ELEMENTS(val));
  val[i] = 0;

  return g_strdup (val);
}
